/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.List;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;

import com.google.inject.Binding;
import com.google.inject.Injector;
import com.google.inject.Provider;
import com.google.inject.TypeLiteral;
import com.google.inject.spi.ElementSource;

/**
 * This {@link MethodInterceptor} is intended to be used with {@link DecoderModule}s where multiple
 * ones "produce" the same {@link PhenotypeDecoder}.
 * It intercepts the decode method call (defined by an annotation) and launches all sub-
 * {@link DecoderModule}s before executing its own decoding. Sub-{@link DecoderModule}s are not
 * allowed to interfere with each other.
 * 
 * @author diewald
 */
public class DecodingPluginInterceptor implements MethodInterceptor {

	/**
	 * Reference to the {@link Injector} used to execute the decoding. Required to find all
	 * {@link DecoderModule}s that produce the same type.
	 */
	protected Injector decodingInjector;

	/** Sets the {@link Injector} used in the {@link DecodingInterceptModule} phase. */
	public void setDecodingInjector(Injector decodingInjector) {
		this.decodingInjector = decodingInjector;
	}

	/** {@inheritDoc} */
	@Override
	public Object invoke(MethodInvocation invocation) throws Throwable {
		Class<?> returnType = invocation.getMethod().getReturnType();
		@SuppressWarnings("unchecked") List<Binding<?>> subDecoderBindings =
				(List<Binding<?>>)(List<?>)decodingInjector.findBindingsByType(TypeLiteral
						.get(returnType));

		for(Binding<?> decodeBinding : subDecoderBindings) {
			Provider<?> provider = decodeBinding.getProvider();
			ElementSource source = (ElementSource)decodeBinding.getSource();
			Object declaringSource = source.getDeclaringSource();
			if(!(declaringSource instanceof Method)) {
				// TODO: Proper error message, modify the exception to take the phenotype instead of
				// the genotype.
				System.out.println("Wrong");
			}
			Method declaringMethod = (Method)declaringSource;
			if(!overrides(invocation.getMethod(), declaringMethod)) {
				provider.get();
			}
			System.out.println("Hi");
		}

		return invocation.proceed();
	}

	/**
	 * Returns true if {@code override} overrides {@code base}, assumes that the signatures match.
	 * Taken from Google Guice.
	 */
	private static boolean overrides(Method override, Method base) {
		// First, check the inheritance hierarchy, then compare the method modifiers and signatures.
		if(!base.getDeclaringClass().isAssignableFrom(override.getDeclaringClass())) {
			return false;
		}

		// See JLS section 8.4.8.1
		int modifiers = base.getModifiers();
		if(Modifier.isPublic(modifiers) || Modifier.isProtected(modifiers)) {
			return true;
		}
		if(Modifier.isPrivate(modifiers)) {
			return false;
		}
		// b must be package-private
		return override.getDeclaringClass().getPackage()
				.equals(base.getDeclaringClass().getPackage());
	}
}
