/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingRuntimeException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.start.Opt4JTask;

import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Provides;

/**
 * {@link Decoder} that handles dependencies between {@link Decoder}s, {@link Genotype}s, and
 * {@link Phenotype}s using the Guice framework.
 * 
 * @author diewald
 */
public class GuiceDecoder extends AbstractModule implements Decoder<Genotype, Phenotype> {

	/** Resulting phenotype from the decoding phase that is composed of sub-{@link Phenotype}s. */
	protected Phenotype composedPhenotype;

	/** Defines the desired exploration result type. */
	// TODO: Remove this hard coded line and get the desired result from the backend interface.
	protected Set<Class<? extends Phenotype>> resultTypes;

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<DecoderModule<?>> enabledDecoders;

	/** {@link Opt4JTask} provider used to obtain the opt4j {@link Injector}. */
	@Inject
	protected Opt4JDseTaskProvider opt4JProvider;

	/** Child {@link Injector} for the dependency management of the decode phase. */
	protected Injector decodingInjector;

	/**
	 * Holds the {@link MethodInterceptor} that is executed when a decode method capable of
	 * launching sub-decoders is hit.
	 */
	protected DecodingPluginInterceptor decodingInterceptor;

	/** Reference to the input {@link Genotype}. */
	// @ThreadScoped
	protected ThreadLocal<CompositeGenotype<Class<? extends Genotype>, ? extends Genotype>> genotype =
			new ThreadLocal<>();

	/** Constructor. Remembers the demanded {@link Phenotype}ss to be produced. */
	public GuiceDecoder(Set<Class<? extends Phenotype>> resultTypes,
			Collection<DecoderModule<?>> enabledDecoders) {
		this.resultTypes = resultTypes;
		this.enabledDecoders = enabledDecoders;
	}

	// To be used later: Constructor injection is favorable.
	// /** Constructor. Gets the Opt4J injector by Guice. */
	// @Inject
	// public GuiceDecoder(Opt4JDseTaskProvider provider) {
	// decodingInjector = provider.get().get().createChildInjector(this);
	// }

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Phenotype decode(Genotype genotype) {
		// The decode operations are responsible to add the results of the decode operation to the
		// parent phenotype.
		if(!(genotype instanceof CompositeGenotype)) {
			throw new DecodingRuntimeException(genotype, this,
					"The top-most DSE process decoder requires a CompositeGenotype as input!");
		}
		this.genotype
				.set((CompositeGenotype<Class<? extends Genotype>, ? extends Genotype>)genotype);
		// TODO(AD): See configure(): We need a service mechanism and user configuration to select
		// the desired set of activated decoders.
		decodingInjector = opt4JProvider.get().get().createChildInjector(this);
		DecodingPluginInterceptor interceptor =
				decodingInjector.getInstance(DecodingPluginInterceptor.class);
		interceptor.setDecodingInjector(decodingInjector);

		// FIXME: Hack to get the correct instance: Is multiprovisioning supported by Guice?
		Class<? extends Phenotype> resultType = getFirst(resultTypes).get();
		decodingInjector.getInstance(resultType);
		return composedPhenotype;
	}

	/** Provides the composite {@link Genotype} as required by the Guice framework. */
	@Provides
	// @ThreadScoped
			public
			CompositeGenotype<Class<? extends Genotype>, ? extends Genotype> getGenotype() {
		return genotype.get();
	}

	/** Returns the sub-genotype of the given type, or {@code null} if none is found. */
	public <T extends Genotype> T getGenotype(Class<T> genotypeType) {
		return getGenotype().get(genotypeType);
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		// bindScope(ThreadScoped.class, new ThreadScope());
		install(this);

		for(DecoderModule<?> module : enabledDecoders) {
			install(module);
		}
	}

	/*
	 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	 * HACKs below! They are used to retrieve the concrete genotypes. Scoping shall resolve the
	 * problem of Genotypes living in parallel.
	 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	 */
	/**
	 * Returns the {@link SafetyFunctionArchEncoding} sub-{@link Genotype} passed to this decoder.
	 * (Testing!!!)
	 */
	// TODO: Move to the encoding to render the approach more modular.
	@Provides
	public SafetyFunctionArchEncoding getSFAEnc() {
		return getGenotype(SafetyFunctionArchEncoding.class);
	}

	/** Returns the {@link PartitionMappingEncoding} from the {@link CompositeGenotype}. */
	@Provides
	@Genotyped
	public PartitionMappingEncoding getPMEnc() {
		return getGenotype(PartitionMappingEncoding.class);
	}

	/** Returns the {@link PlatformCommunicationGraphEncoding} from the {@link CompositeGenotype}. */
	@Provides
	@Genotyped
	public PlatformCommunicationGraphEncoding getPCGEnc() {
		return getGenotype(PlatformCommunicationGraphEncoding.class);
	}

	/**
	 * Returns the {@link AbstractTaskMappingEncoding} sub-{@link Genotype} passed to this decoder.
	 * (Testing!!!)
	 */
	@Provides
	@Genotyped
	public AbstractTaskMappingEncoding getATMEnc() {
		return getGenotype(AbstractTaskMappingEncoding.class);
	}

	/**
	 * Returns the {@link FailSilentTaskMappingEncoding} sub-{@link Genotype} passed to this
	 * decoder. (Testing!!!)
	 */
	@Provides
	@Genotyped
	public FailSilentTaskMappingEncoding getITMEnc() {
		return getGenotype(FailSilentTaskMappingEncoding.class);
	}
}
