/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import static com.google.common.collect.Sets.newHashSet;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.rethrow;

import java.util.Collection;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.DSEFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.ArchitectureExplorationEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.GuiceCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.GuiceDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.GenotypeFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.ArchitectureSolution;
import org.fortiss.af3.exploration.alg.dse.evaluator.ArchitectureEvaluator;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.Sets;
import com.google.inject.Injector;

/**
 * Problem module that setups the main control entities constituting the DSE process exploration. It
 * defines the top-most {@link Creator}s, {@link Decoder}s, and {@link Evaluator}s, which consists
 * of sub-modules. Moreover, this class calculates the set of required decoders to be launched on
 * the basis of the activated {@link IExplorationFeature}s and the demanded solutions.
 * 
 * @author diewald
 */
public class DseProcessProblemModule
		extends
		DseProblemModuleBase<ArchitectureExplorationEncoding, ArchitectureSolution, ArchitectureDecoder> {

	/** To be removed. */
	private DSEFactory dseFactory = DSEFactory.getInstance();

	/** Reference to the DSE-wide {@link Injector} to allow access to the system model parameters. */
	private Injector dseInjector;

	/** Constructor. */
	public DseProcessProblemModule(DseSpecification dse, Injector dseInjector,
			Set<Class<? extends Phenotype>> requestedSolutions) {
		super(dse, null, requestedSolutions);
		this.dseInjector = dseInjector;
		dseFactory.init(null);
	}

	/**
	 * Configure {@link Genotype} creator (execution model dependent), problem decoder and goal
	 * evaluators.
	 */
	@Override
	protected void config() {
		/*
		 * Bind the Creators and the Decoders to the opt4j optimization problem. The evaluators are
		 * added separately.
		 */
		DSEFactory dseFactory = DSEFactory.getInstance();
		dseFactory.init(null);

		ArchitectureEvaluator archEvaluator = dseFactory.createArchitectureEvaluator();

		Collection<IExplorationFeature> enabledFeatures =
				getService(IExplorationFeatureService.class).getExplorationFeatures(
						dse.getFeatures());
		Collection<CreatorModule<?>> enabledCreators = newHashSet();
		enabledFeatures.forEach(f -> enabledCreators.addAll(f.getCreators()));
		Collection<DecoderModule<?>> enabledDecoders = newHashSet();
		enabledFeatures.forEach(f -> enabledDecoders.addAll(f.getDecoders()));

		// TODO: construct the execution dependency graph elsewhere?
		// bindProblem(dseFactory.createArchitectureExplorationCreator(),
		// dseFactory.createArchitectureDecoder(execDepGraph), archEvaluator);

		GuiceDecoder guiceDecoder = createGuiceDecoder(enabledDecoders);
		Set<Class<? extends Genotype>> requiredGenotypes = findRequiredGenotypes(guiceDecoder);
		bindProblem(new GuiceCreator(requiredGenotypes, enabledCreators), guiceDecoder,
				archEvaluator);

		// Create/Bind the subproblems of the architectural exploration problem.
		try {
			// dseFactory.createSubProblems(this, dse, execDepGraph);
		} catch(Exception e) {
			throw new RuntimeException("Failed setting up the DSE Problem instance.", e);
		}

		// Create evaluators wrappers (based on provided goal specification)
		try {
			addGoalEvaluators(archEvaluator);
			// addGoalEvaluator(new EnergyObjectiveEvaluator(),
			// PredefinedFactory.eINSTANCE.createEnergyMinObjective(), 0);
		} catch(Exception e) {
			throw rethrow(e);
		}
	}

	/**
	 * Creates a {@link GuiceDecoder}, prepares all {@link DecoderModule}s, and defines the
	 * {@link MethodInterceptor} for plugin-type {@link DecoderModule}s.
	 * 
	 * @param enabledDecoders
	 *            Set of {@link DecoderModule}s configure for this DSE execution.
	 * @return The created/configured {@link GuiceDecoder}.
	 */
	private GuiceDecoder createGuiceDecoder(Collection<DecoderModule<?>> enabledDecoders) {
		Collection<DecoderModule<?>> injectedDecoders = Sets.newHashSet();
		for(DecoderModule<?> decoderModule : enabledDecoders) {
			DecoderModule<?> currInjectedDecoderModule =
					dseInjector.getInstance(decoderModule.getClass());
			injectedDecoders.add(currInjectedDecoderModule);
		}
		return new GuiceDecoder(requestedSolutions, injectedDecoders);
	}

	/**
	 * Returns the collection of {@link Genotype}s that are required as inputs to decoders to obtain
	 * the requested {@link Phenotype}.
	 */
	private Set<Class<? extends Genotype>> findRequiredGenotypes(GuiceDecoder guiceDecoder) {
		Injector decodingInjector = dseInjector.createChildInjector(guiceDecoder);
		GenotypeFinder genotypeFinder = new GenotypeFinder();
		return genotypeFinder.getGenotypesFor(requestedSolutions, decodingInjector);
	}
}
