/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultListenableGraph;
import org.jgrapht.graph.DirectedWeightedMultigraph;

/**
 * Interface for {@link TaskGraph}s which are the (flattened) DSE-internal representations of a
 * (hierarchical) logical architecture.
 * 
 * @author diewald
 */
public interface TaskGraph {
	// TODO: Get rid of this method!
	/** Returns the underlying {@link DefaultListenableGraph}. */
	DefaultListenableGraph<ITaskAdapter<?>, SignalAdapterWeightedEdge> getGraph();

	// TODO: Get rid of this method!
	/**
	 * Returns the underlying {@link DirectedWeightedMultigraph} that holds the data of the
	 * corresponding {@link DefaultListenableGraph}.
	 */
	DirectedWeightedMultigraph<ITaskAdapter<?>, SignalAdapterWeightedEdge> getDelegateGraph();

	/**
	 * Updates the period of this {@link TaskGraph}: Determine the common period of the tasks
	 * within
	 * an independent task graph if it exists. Otherwise, set it to 0.
	 * 
	 * @throws InvalidTimingModelException
	 *             if the periods of the {@link TaskGraph}'s task are inconsistent.
	 */
	void updatePeriod() throws InvalidTimingModelException;

	/** Returns the {@link TaskGraph}s period. */
	Double getPeriod();

	/** Returns the set of deployable components (=tasks) of this {@link TaskGraph}. */
	Set<ITaskAdapter<?>> getTasks();

	/**
	 * Checks whether {@code this} {@link TaskGraph} contains the given {@link ITaskAdapter} or an
	 * instance of it.
	 */
	boolean containsTask(ITaskAdapter<?> task);

	/**
	 * Checks whether {@code this} {@link TaskGraph} contains any of the given {@link ITaskAdapter}
	 * s.
	 */
	boolean containsAnyTaskOf(Collection<ITaskAdapter<?>> taskSet);

	/**
	 * Adds the given {@link ITaskAdapter} to the {@link TaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}.
	 */
	boolean addTask(ITaskAdapter<?> task);

	/**
	 * Adds the given {@code task} to the {@link TaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}, and further makes it accessible if the given
	 * {@code refTask} is queried.
	 */
	boolean addTask(ITaskAdapter<?> task, ITaskAdapter<?> refTask);

	/**
	 * Adds all given tasks to the task graph.
	 * 
	 * @param tasks
	 *            collection of tasks to add to the graph.
	 * @return whether the tasks were successfully added.
	 */
	boolean addAllTask(Collection<ITaskAdapter<?>> tasks);

	/** Adds a signal to the task graph that defines a data exchange between two tasks. */
	boolean addSignal(ISignalAdapter<?, ?> signalAdapter);

	/** Adds signals to the task graph that are used to exchange data between tasks. */
	boolean addAllSignals(Collection<ISignalAdapter<?, ?>> signalAdapters);

	/** Removes the given {@code task} and its data dependencies from the {@link TaskGraph}. */
	void removeTask(ITaskAdapter<?> task);

	/** Creates a Dialog that illustrates this {@link TaskGraph}. */
	void display();

	/** Creates and returns a copy of {@code this} {@link TaskGraph}. */
	<T extends TaskGraph> T copy();

	/** Registers a fault detector for a given task. */
	void addFaultDetector(ITaskAdapter<?> deployableComponent, FaultDetector faultDetector);

	/** Returns the fault detector corresponding to the given Task and ID. */
	// TODO: Remove, bad design. Fault detectors are tasks and should be added in a decoder to
	// another phenotype.
	FaultDetector getFaultDetector(ITaskAdapter<?> task, int id);

	/** Returns a random fault detector index / ID for the given task. */
	int getRandomFaultDetectorId(ITaskAdapter<?> deployableComponent);
}
